searchNodes=[{"doc":"Broadway is a concurrent, multi-stage tool for building data ingestion and data processing pipelines. It allows developers to consume data efficiently from different sources, such as Amazon SQS, Apache Kafka, Google Cloud PubSub, RabbitMQ and others. Built-in features Back-pressure - by relying on GenStage , we only get the amount of events necessary from upstream sources, never flooding the pipeline. Automatic acknowledgements - Broadway automatically acknowledges messages at the end of the pipeline or in case of errors. Batching - Broadway provides built-in batching, allowing you to group messages either by size and/or by time. This is important in systems such as Amazon SQS, where batching is the most efficient way to consume messages, both in terms of time and cost. Fault tolerance - Broadway pipelines are carefully designed to manage failures. Producers are isolated from the rest of the pipeline and automatically resubscribe to your data source in case of crashes. At the same time, all of your Broadway callbacks are stateless, which allows Broadway to handle any errors locally. This provides a stable foundation that play well with your producers, regardless if their delivery guarantees are at least once, at most once, or exactly once. Graceful shutdown - Broadway integrates with the VM to provide graceful shutdown. By starting Broadway as part of your supervision tree, it will guarantee all events are flushed once the VM shuts down. Built-in testing - Broadway ships with a built-in test API, making it easy to push test messages through the pipeline and making sure the event was properly processed. Custom failure handling - Broadway provides a handle_failed/2 callback where developers can outline custom code to handle errors. For example, if they want to move messages to another queue for further processing. Dynamic batching - Broadway allows developers to batch messages based on custom criteria. For example, if your pipeline needs to build batches based on the user_id , email address, etc, it can be done by calling Broadway.Message.put_batch_key/2 . Ordering and Partitioning - Broadway allows developers to partition messages across workers, guaranteeing messages within the same partition are processed in order. For example, if you want to guarantee all events tied to a given user_id are processed in order and not concurrently, you can set the :partition_by option. See &quot;Ordering and partitioning&quot; . Rate limiting: Broadway allows developers to rate limit all producers in a single node by a given number of messages in a time period, allowing developers to easily work sources or sinks that cannot cope with a high number of requests. See the &quot;:rate_limiting&quot; option for producers in start_link/2 . Metrics - Broadway uses the :telemetry library for instrumentation, see &quot;Telemetry&quot; section below for more information. Back-off (TODO) The Broadway Behaviour In order to use Broadway, you need to: Define your pipeline configuration Define a module implementing the Broadway behaviour Example Broadway is a process-based behaviour, and you begin by defining a module that invokes use Broadway . Processes defined by these modules will often be started by a supervisor, and so a start_link/1 function is frequently also defined but not strictly necessary. defmodule MyBroadway do use Broadway def start_link ( _opts ) do Broadway . start_link ( MyBroadway , name : MyBroadwayExample , producer : [ module : { Counter , [ ] } , concurrency : 1 ] , processors : [ default : [ concurrency : 2 ] ] ) end ... callbacks ... end Then add your Broadway pipeline to your supervision tree (usually in lib/my_app/application.ex ): children = [ { MyBroadway , [ ] } ] Supervisor . start_link ( children , strategy : :one_for_one ) Adding your pipeline to your supervision tree in this way calls the default child_spec/1 function that is generated when use Broadway is invoked. If you would like to customize the child spec passed to the supervisor, you can override the child_spec/1 function in your module or explicitly pass a child spec to the supervisor when adding it to your supervision tree. The configuration above defines a pipeline with: One producer Two processors Here is how this pipeline would be represented: [producer_1] / \\ / \\ / \\ / \\ [processor_1] [processor_2] &lt;- process each message After the pipeline is defined, you need to implement the handle_message/3 callback which will be invoked by processors for each message. handle_message/3 receives every message as a Broadway.Message struct and it must return an updated message. Batching Depending on the scenario, you may want to group processed messages as batches before publishing your data. This is common and especially important when working with services like AWS S3 and SQS that provide a specific API for sending and retrieving batches. This can drastically increase throughput and consequently improve the overall performance of your pipeline. To create batches, define the :batchers configuration option: defmodule MyBroadway do use Broadway def start_link ( _opts ) do Broadway . start_link ( MyBroadway , name : MyBroadwayExample , producer : [ module : { Counter , [ ] } , concurrency : 1 ] , processors : [ default : [ concurrency : 2 ] ] , batchers : [ sqs : [ concurrency : 2 , batch_size : 10 ] , s3 : [ concurrency : 1 , batch_size : 10 ] ] ) end # ...callbacks... end The configuration above defines a pipeline with: One producer Two processors One batcher named :sqs with two batch processors One batcher named :s3 with one batch processor Here is how this pipeline would be represented: [producer_1] / \\ / \\ / \\ / \\ [processor_1] [processor_2] &lt;- process each message /\\ /\\ / \\ / \\ / \\ / \\ / x \\ / / \\ \\ / / \\ \\ / / \\ \\ [batcher_sqs] [batcher_s3] /\\ \\ / \\ \\ / \\ \\ / \\ \\ [batch_sqs_1] [batch_sqs_2] [batch_s3_1] &lt;- process each batch Additionally, you have to define the handle_batch/4 callback, which batch processors invoke for each batch. You can then call Broadway.Message.put_batcher/2 inside handle_message/3 to control which batcher the message should go to. The batcher receives processed messages and creates batches specified by the batch_size and batch_timeout configuration. The goal is to create a batch with at most batch_size entries within batch_timeout milliseconds. Each message goes into a particular batch, controlled by calling Broadway.Message.put_batch_key/2 in handle_message/3 . Once a batch is created in the batcher, it is sent to a separate process (the batch processor) that will call handle_batch/4 , passing the batcher, the batch itself (a list of messages), a Broadway.BatchInfo struct, and the Broadway context. For example, imagine your producer generates integers as data . You want to route the odd integers to SQS and the even ones to S3. Your pipeline would look like this: defmodule MyBroadway do use Broadway import Integer alias Broadway.Message # ...start_link... @impl true def handle_message ( _ , % Message { data : data } = message , _ ) when is_odd ( data ) do message |&gt; Message . update_data ( &amp; process_data / 1 ) |&gt; Message . put_batcher ( :sqs ) end def handle_message ( _ , % Message { data : data } = message , _ ) when is_even ( data ) do message |&gt; Message . update_data ( &amp; process_data / 1 ) |&gt; Message . put_batcher ( :s3 ) end defp process_data ( data ) do # Do some calculations, generate a JSON representation, etc. end @impl true def handle_batch ( :sqs , messages , _batch_info , _context ) do # Send batch of successful messages as ACKs to SQS # This tells SQS that this list of messages were successfully processed end def handle_batch ( :s3 , messages , _batch_info , _context ) do # Send batch of messages to S3 end end See the callbacks documentation for more information on the arguments given to each callback and their expected return types. The default batcher Once you define the :batchers configuration key for your Broadway pipeline, then all messages get batched . By default, unless you call Broadway.Message.put_batcher/2 , messages have their batcher set to the :default batcher. If you don't define configuration for it, Broadway is going to raise an error. For example, imagine you want to batch &quot;special&quot; messages and handle them differently then all other messages. You can configure your pipeline like this: defmodule MyBroadway do use Broadway def start_link ( _opts ) do Broadway . start_link ( MyBroadway , name : MyBroadwayExample , producer : [ module : { Counter , [ ] } , concurrency : 1 ] , processors : [ default : [ concurrency : 2 ] ] , batchers : [ special : [ concurrency : 2 , batch_size : 10 ] , default : [ concurrency : 1 , batch_size : 10 ] ] ) end def handle_message ( _ , message , _ ) do if special? ( message ) do Broadway.Message . put_batcher ( :special ) else message end end def handle_batch ( :special , messages , _batch_info , _context ) do # Handle special batch end def handle_batch ( :default , messages , _batch_info , _context ) do # Handle all other messages in batches end Now you are ready to get started. See the start_link/2 function for a complete reference on the arguments and options allowed. Also makes sure to check out GUIDES in the documentation sidebar for more examples, how tos and more. Acknowledgements and failures At the end of the pipeline, messages are automatically acknowledged. If there are no batchers, the acknowledgement will be done by processors. The number of messages acknowledged, assuming the pipeline is running at full scale, will be max_demand - min_demand . Since the default values are 10 and 5 respectively, we will be acknowledging in groups of 5. If there are batchers, the acknowledgement is done by the batchers, using the batch_size . In case of failures, Broadway does its best to keep the failures contained and avoid losing messages. The failed message or batch is acknowledged as failed immediately. For every failure, a log report is also emitted. If your Broadway module also defines the handle_failed/2 callback, that callback will be invoked with all the failed messages before they get acknowledged. Note however, that Broadway does not provide any sort of retries out of the box. This is left completely as a responsibility of the producer. For instance, if you are using Amazon SQS, the default behaviour is to retry unacknowledged messages after a user-defined timeout. If you don't want unacknowledged messages to be retried, is your responsibility to configure a dead-letter queue as target for those messages. Producer concurrency Setting producer concurrency is a tradeoff between latency and internal queueing. For efficiency, you should generally limit the amount of internal queueing. Whenever additional messages are sitting in a busy processor's mailbox, they can't be delivered to another processor which may be available or become available first. One possible cause of internal queueing is multiple producers. This is because each processor's demand will be sent to all producers. For example, if a processor demands 2 messages and there are 2 producers, each producer will try to produce 2 messages (for example, by pulling from a queue or whatever the specific producer does) and give them to the processor. So the processor may receive max_demand * &lt;producer concurrency&gt; messages. Setting producer concurrency: 1 will reduce internal queueing. This is likely a good choice for producers which take minimal time to produce a message, such as BroadwayRabbitMQ , which receives messages as they are pushed by RabbitMQ and can specify how many to prefetch. On the other hand, when using a producer such as BroadwaySQS which must make a network round trip to fetch from an external source, it may be better to use multiple producers and accept some internal queueing to avoid having fetch messages whenever there is new demand. Measure your system to decide which setting is most appropriate. Adding another single-producer pipeline, or another node running the pipeline, are other ways you may consider to increase throughput. Batcher concurrency If a batcher's concurrency is greater than 1 , Broadway will use as few of the batcher processes as possible at any given moment, attempting to satisfy the batch_size of one batcher process within the batch_timeout before sending messages to another. Testing Many producers receive data from external systems and hitting the network is usually undesirable when running the tests. For testing purposes, we recommend developers to use Broadway.DummyProducer . This producer does not produce any messages by itself and instead the test_message/3 and test_batch/3 functions should be used to publish messages. With test_message/3 , you can push a message into the pipeline and receive a process message when the pipeline acknowledges the data you have pushed has been processed. Let's see an example. Imagine the following Broadway module: defmodule MyBroadway do use Broadway def start_link ( ) do producer_module = Application . fetch_env! ( :my_app , :producer_module ) Broadway . start_link ( __MODULE__ , name : __MODULE__ , producer : [ module : producer_module ] , processors : [ default : [ ] ] , batchers : [ default : [ batch_size : 10 ] ] ) end @impl true def handle_message ( _processor , message , _context ) do message end @impl true def handle_batch ( _batcher , messages , _batch_info , _context ) do messages end end Now in config/test.exs you could do: config :my_app , :producer_module , { Broadway.DummyProducer , [ ] } And we can test it like this: defmodule MyBroadwayTest do use ExUnit.Case , async : true test &quot;test message&quot; do ref = Broadway . test_message ( MyBroadway , 1 ) assert_receive { :ack , ^ ref , [ %{ data : 1 } ] , [ ] } end end Note that at the end we received a message in the format of: { :ack , ^ ref , successful_messages , failure_messages } You can use the acknowledgment to guarantee the message has been processed and therefore any side-effect from the pipeline should be visible. When using test_message/3 , the message will be delivered as soon as possible, without waiting for the pipeline batch_size to be reached or without waiting for batch_timeout . This behaviour is useful to test and verify single messages, without imposing high timeouts to our test suites. In case you want to test multiple messages, then you need to use test_batch/3 . test_batch/3 will respect the batching configuration, which most likely means you need to increase your test timeouts: test &quot;batch messages&quot; do { :ok , pid } = MyBroadway . start_link ( ) ref = Broadway . test_batch ( pid , [ 1 , 2 , 3 ] ) assert_receive { :ack , ^ ref , [ %{ data : 1 } , %{ data : 2 } , %{ data : 3 } ] , [ ] } , 1000 end However, keep in mind that, generally speaking, there is no guarantee the messages will arrive in the same order that you have sent them, especially for large batches, as Broadway will process large batches concurrently and order will be lost. If you want to send more than one test message at once, then we recommend setting the :batch_mode to :bulk , especially if you want to assert how the code will behave with large batches. Otherwise the batcher will flush messages as soon as possible and in small batches. However, keep in mind that, regardless of the :batch_mode you cannot rely on ordering, as Broadway pipelines are inherently concurrent. For example, if you send those messages: test &quot;multiple batch messages&quot; do { :ok , pid } = MyBroadway . start_link ( ) ref = Broadway . test_batch ( pid , [ 1 , 2 , 3 , 4 , 5 , 6 , 7 ] , batch_mode : :bulk ) assert_receive { :ack , ^ ref , [ %{ data : 1 } ] , [ ] } , 1000 end Testing with Ecto If you are using Ecto in your Broadway processors and you want to run your tests concurrently, you need to tell Broadway to use the Ecto SQL Sandbox during tests. This can be done in two steps. First, when you call test_messages/3 in your tests, include the :ecto_sandbox process in the message metadata: Broadway . test_message ( MyApp.Pipeline , message , metadata : %{ ecto_sandbox : self ( ) } ) Now we can use Broadway telemetry callbacks to fetch the sandbox process and enable it inside the processor. Add to your test/test_helper.exs : defmodule BroadwayEctoSandbox do def attach ( Repo ) do events = [ [ :broadway , :processor , :start ] , [ :broadway , :batch_processor , :start ] , ] :telemetry . attach_many ( { __MODULE__ , repo } , events , &amp; handle_event / 4 , %{ repo : repo } ) end def handle_event ( _event_name , _event_measurement , %{ messages : messages } , %{ repo : repo } ) do with [ % Broadway.Message { metadata : %{ ecto_sandbox : pid } } | _ ] &lt;- messages do Ecto.Adapters.SQL.Sandbox . allow ( repo , pid , self ( ) ) end :ok end end BroadwayEctoSandbox . attach ( MyApp.Repo ) And now you should have concurrent Broadway tests that talk to the database. Ordering and partitioning By default, Broadway processes all messages and batches concurrently, which means ordering is not guaranteed. Some producers may impose some ordering (for instance, Apache Kafka), but if the ordering comes from a business requirement, you will have to impose the ordering yourself. This can be done with the :partition_by option, which enforces that messages with a given property are always forwarded to the same stage. In order to provide partitioning throughout the whole pipeline, just set :partition_by at the root of your configuration: defmodule MyBroadway do use Broadway def start_link ( _opts ) do Broadway . start_link ( MyBroadway , name : MyBroadwayExample , producer : [ module : { Counter , [ ] } , concurrency : 1 ] , processors : [ default : [ concurrency : 2 ] ] , batchers : [ sqs : [ concurrency : 2 , batch_size : 10 ] , s3 : [ concurrency : 1 , batch_size : 10 ] ] , partition_by : &amp; partition / 1 ) end defp partition ( msg ) do msg . data . user_id end In the example above, we are partitioning the pipeline by user_id . This means any message with the same user_id will be handled by the same processor and batch processor. The partition function must return a non-negative integer, starting at zero, which is routed to a stage by using the remainder option. If the data you want to partition by is not an integer, you can explicitly hash it by calling :erlang.phash2/1 . However, note that hash does not guarantee an equal distribution of events across partitions. So some partitions may be more overloaded than others, slowing down the whole pipeline. In the example above, we have set the same partition for all processors and batchers. You can also specify the :partition_by function for each &quot;processor&quot; and &quot;batcher&quot; individually. Finally, beware of the error semantics when using partitioning. If you require ordering and a message fails, the partition will continue processing messages. Depending on the type of processing, the end result may be inconsistent. If your producer supports retrying, the failed message may be retried later, also out of order. Those issues happens regardless of Broadway and solutions to said problems almost always need to be addressed outside of Broadway too. Telemetry Broadway currently exposes following Telemetry events: [:broadway, :topology, :init] - Dispatched when the topology for a Broadway pipeline is initialized. The config key in the metadata contains the configuration options that were provided to Broadway.start_link/2 . Measurement: %{system_time: integer} Metadata: %{supervision: pid, config: keyword} [:broadway, :processor, :start] - Dispatched by a Broadway processor before the optional prepare_messages/2 Measurement: %{system_time: integer} Metadata: %{ topology_name : atom , name : atom , processor_key : atom , index : non_neg_integer , messages : [ Broadway.Message . t ] , telemetry_span_context : reference } [:broadway, :processor, :stop] - Dispatched by a Broadway processor after prepare_messages/2 and after all handle_message/3 callback has been invoked for all individual messages Measurement: %{duration: native_time} Metadata: %{ topology_name : atom , name : atom , processor_key : atom , index : non_neg_integer , successful_messages_to_ack : [ Broadway.Message . t ] , successful_messages_to_forward : [ Broadway.Message . t ] , failed_messages : [ Broadway.Message . t ] , telemetry_span_context : reference } [:broadway, :processor, :message, :start] - Dispatched by a Broadway processor before your handle_message/3 callback is invoked Measurement: %{system_time: integer} Metadata: %{ processor_key : atom , topology_name : atom , name : atom , index : non_neg_integer , message : Broadway.Message . t , telemetry_span_context : reference } [:broadway, :processor, :message, :stop] - Dispatched by a Broadway processor after your handle_message/3 callback has returned Measurement: %{duration: native_time} Metadata: %{ processor_key : atom , topology_name : atom , name : atom , index : non_neg_integer , message : Broadway.Message . t , telemetry_span_context : reference } [:broadway, :processor, :message, :exception] - Dispatched by a Broadway processor if your handle_message/3 callback encounters an exception Measurement: %{duration: native_time} Metadata: %{ processor_key : atom , topology_name : atom , name : atom , index : non_neg_integer , message : Broadway.Message . t , kind : kind , reason : reason , stacktrace : stacktrace , telemetry_span_context : reference } [:broadway, :batch_processor, :start] - Dispatched by a Broadway batch processor before your handle_batch/4 callback is invoked Measurement: %{system_time: integer} Metadata: %{ topology_name : atom , name : atom , index : non_neg_integer , messages : [ Broadway.Message . t ] , batch_info : Broadway.BatchInfo . t , telemetry_span_context : reference } [:broadway, :batch_processor, :stop] - Dispatched by a Broadway batch processor after your handle_batch/4 callback has returned Measurement: %{duration: native_time} Metadata: %{ topology_name : atom , name : atom , index : non_neg_integer , successful_messages : [ Broadway.Message . t ] , failed_messages : [ Broadway.Message . t ] , batch_info : Broadway.BatchInfo . t , telemetry_span_context : reference } [:broadway, :batcher, :start] - Dispatched by a Broadway batcher before handling events Measurement: %{system_time: integer} Metadata: %{ topology_name : atom , name : atom , batcher_key : atom , messages : [ Broadway.Message . t ] , telemetry_span_context : reference } [:broadway, :batcher, :stop] - Dispatched by a Broadway batcher after handling events Measurement: %{duration: native_time} Metadata: %{ topology_name : atom , name : atom , batcher_key : atom , telemetry_span_context : reference } Most of the events follow the :telemetry.span/3 convention for measurements. This means that &quot;start&quot; events have a :system_time representing the start of that event using System.system_time/0 . The &quot;stop&quot; or &quot;exception&quot; events have the duration value, which is the difference in monotonic time between the start and stop events.","ref":"Broadway.html","title":"Broadway","type":"behaviour"},{"doc":"Returns all running Broadway names. It's important to notice that no order is guaranteed.","ref":"Broadway.html#all_running/0","title":"Broadway.all_running/0","type":"function"},{"doc":"Gets the current values used for the producer rate limiting of the given pipeline. Returns {:ok, info} if rate limiting is enabled for the given pipeline or {:error, reason} if the given pipeline doesn't have rate limiting enabled. The returned info is a map with the following keys: :interval :allowed_messages See the :rate_limiting options in the module documentation for more information. Examples Broadway . get_rate_limiting ( broadway ) #=&gt; {:ok, %{allowed_messages: 2000, interval: 1000}}","ref":"Broadway.html#get_rate_limiting/1","title":"Broadway.get_rate_limiting/1","type":"function"},{"doc":"Invoked to handle generated batches. It expects: batcher is the key that defined the batcher. This value can be set in the handle_message/3 callback using Broadway.Message.put_batcher/2 . messages is the list of Broadway.Message structs in the incoming batch. batch_info is a Broadway.BatchInfo struct containing extra information about the incoming batch. context is the user defined data structure passed to start_link/2 . It must return an updated list of messages. All messages received must be returned, otherwise an error will be logged. All messages after this step will be acknowledged according to their status. In case of errors in this callback, the error will be logged and the whole batch will be failed. This callback also traps exits, so failures due to broken links between processes do not automatically cascade. For more information on batching, see the &quot;Batching&quot; section in the Broadway documentation.","ref":"Broadway.html#c:handle_batch/4","title":"Broadway.handle_batch/4","type":"callback"},{"doc":"Invoked for failed messages (if defined). It expects: messages is the list of messages that failed. If a message is failed in handle_message/3 , this will be a list with a single message in it. If some messages are failed in handle_batch/4 , this will be the list of failed messages. context is the user-defined data structure passed to start_link/2 . This callback must return the same messages given to it, possibly updated. For example, you could update the message data or use Broadway.Message.configure_ack/2 in a centralized place to configure how to ack the message based on the failure reason. This callback is optional. If present, it's called before the messages are acknowledged according to the producer. This gives you a chance to do something with the message before it's acknowledged, such as storing it in an external persistence layer or similar. This callback is also invoked if handle_message/3 or handle_batch/4 crash or raise an error. If this callback crashes or raises an error, the messages are failed internally by Broadway to avoid crashing the process.","ref":"Broadway.html#c:handle_failed/2","title":"Broadway.handle_failed/2","type":"callback"},{"doc":"Invoked to handle/process individual messages sent from a producer. It receives: processor is the key that defined the processor. message is the Broadway.Message struct to be processed. context is the user defined data structure passed to start_link/2 . And it must return the (potentially) updated Broadway.Message struct. This is the place to do any kind of processing with the incoming message, e.g., transform the data into another data structure, call specific business logic to do calculations. Basically, any CPU bounded task that runs against a single message should be processed here. In order to update the data after processing, use the Broadway.Message.update_data/2 function. This way the new message can be properly forwarded and handled by the batcher: @impl true def handle_message ( _ , message , _ ) do message |&gt; update_data ( &amp; do_calculation_and_returns_the_new_data / 1 ) end In case more than one batcher have been defined in the configuration, you need to specify which of them the resulting message will be forwarded to. You can do this by calling put_batcher/2 and returning the new updated message: @impl true def handle_message ( _ , message , _ ) do # Do whatever you need with the data ... message |&gt; put_batcher ( :s3 ) end Any message that has not been explicitly failed will be forwarded to the next step in the pipeline. If there are no extra steps, it will be automatically acknowledged. In case of errors in this callback, the error will be logged and that particular message will be immediately acknowledged as failed, not proceeding to the next steps of the pipeline. This callback also traps exits, so failures due to broken links between processes do not automatically cascade.","ref":"Broadway.html#c:handle_message/3","title":"Broadway.handle_message/3","type":"callback"},{"doc":"Invoked for preparing messages before handling (if defined). It expects: message is the Broadway.Message struct to be processed. context is the user defined data structure passed to start_link/2 . This is the place to prepare and preload any information that will be used by handle_message/3 . For example, if you need to query the database, instead of doing it once per message, you can do it on this callback. The length of the list of messages received by this callback is based on the min_demand / max_demand configuration in the processor. This callback must always return all messages it receives, as handle_message/3 is still called individually for each message afterwards.","ref":"Broadway.html#c:prepare_messages/2","title":"Broadway.prepare_messages/2","type":"callback"},{"doc":"Returns the names of producers. Examples iex&gt; Broadway . producer_names ( MyBroadway ) [ MyBroadway.Producer_0 , MyBroadway.Producer_1 , ... , MyBroadway.Producer_7 ]","ref":"Broadway.html#producer_names/1","title":"Broadway.producer_names/1","type":"function"},{"doc":"Sends a list of Broadway.Message s to the Broadway pipeline. The producer is randomly chosen among all sets of producers/stages. This is used to send out of band data to a Broadway pipeline.","ref":"Broadway.html#push_messages/2","title":"Broadway.push_messages/2","type":"function"},{"doc":"Starts a Broadway process linked to the current process. module is the module implementing the Broadway behaviour. Options In order to set up how the pipeline created by Broadway should work, you need to specify the blueprint of the pipeline. You can do this by passing a set of options to start_link/2 . Each component of the pipeline has its own set of options. The broadway options are: :name - Required. Used for name registration. When an atom, all processes/stages created will be named using this value as prefix. :shutdown - Optional. The time in milliseconds given for Broadway to gracefully shutdown without discarding events. Defaults to 30_000 (millisecond). The default value is 30000 . :max_restarts - The default value is 3 . :max_seconds - The default value is 5 . :resubscribe_interval - The interval in milliseconds that processors wait until they resubscribe to a failed producers. Defaults to 100 (millisecond). The default value is 100 . :context - A user defined data structure that will be passed to handle_message/3 and handle_batch/4. The default value is :context_not_set . :producer - Required. A keyword list of options. See &quot;Producers options&quot; section below. Only a single producer is allowed. :processors - Required. A keyword list of named processors where the key is an atom as identifier and the value is another keyword list of options. See &quot;Processors options&quot; section below. Currently only a single processor is allowed. :batchers - A keyword list of named batchers where the key is an atom as identifier and the value is another keyword list of options. See &quot;Batchers options&quot; section below. The default value is [] . :partition_by - A function that controls how data is partitioned across all processors and batchers. It receives a Broadway.Message and it must return a non-negative integer, starting with zero, that will be mapped to one of the existing processors. See &quot;Ordering and Partitioning&quot; in the module docs for more information. :spawn_opt - Low-level options given when starting a process. Applies to producers, processors, and batchers. See erlang:spawn_opt/2 for more information. :hibernate_after - If a process does not receive any message within this interval, it will hibernate, compacting memory. Applies to producers, processors, and batchers. Defaults to 15_000 (millisecond). The default value is 15000 . Producers options The producer options allow users to set up the producer. The available options are: :module - Required. A tuple representing a GenStage producer. The tuple format should be {mod, arg} , where mod is the module that implements the GenStage behaviour and arg the argument that will be passed to the init/1 callback of the producer. See Broadway.Producer for more information. :concurrency - The number of concurrent producers that will be started by Broadway. Use this option to control the concurrency level of each set of producers. The default value is 1 . The default value is 1 . :transformer - A tuple representing a transformer that translates a produced GenStage event into a %Broadway.Message{} . The tuple format should be {mod, fun, opts} and the function should have the following spec (event :: term, opts :: term) :: Broadway.Message.t This function must be used sparingly and exclusively to convert regular messages into Broadway.Message . That's because a failure in the :transformer callback will cause the whole producer to terminate, possibly leaving unacknowledged messages along the way. The default value is nil . :spawn_opt - Overrides the top-level :spawn_opt . :hibernate_after - Overrides the top-level :hibernate_after . :rate_limiting - A list of options to enable and configure rate limiting for producing. If this option is present, rate limiting is enabled, otherwise it isn't. Rate limiting refers to the rate at which producers will forward messages to the rest of the pipeline. The rate limiting is applied to and shared by all producers within the time limit. The following options are supported: :allowed_messages - Required. An integer that describes how many messages are allowed in the specified interval. :interval - Required. An integer that describes the interval (in milliseconds) during which the number of allowed messages is allowed. If the producer produces more than allowed_messages in interval , only allowed_messages will be published until the end of interval , and then more messages will be published. Processors options :concurrency - The number of concurrent process that will be started by Broadway. Use this option to control the concurrency level of the processors. The default value is System.schedulers_online() * 2 . :min_demand - Set the minimum demand of all processors stages. Default value is 5 . :max_demand - Set the maximum demand of all processors stages. Default value is 10 . The default value is 10 . :partition_by - Overrides the top-level :partition_by . :spawn_opt - Overrides the top-level :spawn_opt . :hibernate_after - Overrides the top-level :hibernate_after . Batchers options :concurrency - The number of concurrent batch processors that will be started by Broadway. Use this option to control the concurrency level. Note that this only sets the numbers of batch processors for each batcher group, not the number of batchers. The number of batchers will always be one for each batcher key defined. The default value is 1 . The default value is 1 . :batch_size - The size of the generated batches. Default value is 100 . The default value is 100 . :batch_timeout - The time, in milliseconds, that the batcher waits before flushing the list of messages. When this timeout is reached, a new batch is generated and sent downstream, no matter if the :batch_size has been reached or not. Default value is 1000 (1 second). The default value is 1000 . :partition_by - Optional. Overrides the top-level :partition_by . :spawn_opt - Overrides the top-level :spawn_opt . :hibernate_after - Overrides the top-level :hibernate_after .","ref":"Broadway.html#start_link/2","title":"Broadway.start_link/2","type":"function"},{"doc":"Synchronously stops the Broadway pipeline with the given reason . This function returns :ok if the server terminates with the given reason; if it terminates with another reason, the call exits. This function keeps OTP semantics regarding error reporting. If the reason is any other than :normal , :shutdown or {:shutdown, _} , an error report is logged.","ref":"Broadway.html#stop/3","title":"Broadway.stop/3","type":"function"},{"doc":"Sends a list of data as a batch of messages to the Broadway pipeline. This is a convenience used for testing. Each message is automatically wrapped in a Broadway.Message with Broadway.CallerAcknowledger configured to send a message back to the caller once all batches have been fully processed. If there are more messages in the batch than the pipeline batch_size or if the messages in the batch take more time to process than batch_timeout then the caller will receive multiple messages. It returns a reference that can be used to identify the ack messages. See &quot;Testing&quot; section in module documentation for more information. Options :metadata - an enumerable of key-value pairs of additional fields to add to the message. This can be used, for example, when testing BroadwayRabbitMQ.Producer . The default value is [] . :acknowledger - a function that generates ack fields of the sent Broadway.Message.t() . This function receives the acknowledger data and the from field and it must return the acknowledger tuple. The typespec of this function is: data :: term(), from :: {pid(), term()} -&gt; {module(), ack_ref :: term(), ack_data :: term()} Examples For example, in your tests, you may do: ref = Broadway . test_batch ( broadway , [ 1 , 2 , 3 ] ) assert_receive { :ack , ^ ref , successful , failed } , 1000 assert length ( successful ) == 3 assert length ( failed ) == 0 Note that messages sent using this function will ignore demand and :transform option specified in :producer option in Broadway.start_link/2 .","ref":"Broadway.html#test_batch/3","title":"Broadway.test_batch/3","type":"function"},{"doc":"Sends a test message through the Broadway pipeline. This is a convenience used for testing. The given data is automatically wrapped in a Broadway.Message with Broadway.CallerAcknowledger configured to send a message back to the caller once the message has been fully processed. The message is set to be flushed immediately, without waiting for the Broadway pipeline batch_size to be filled or the batch_timeout to be triggered. It returns a reference that can be used to identify the ack messages. See &quot;Testing&quot; section in module documentation for more information. Options :metadata - an enumerable of key-value pairs of additional fields to add to the message. This can be used, for example, when testing BroadwayRabbitMQ.Producer . The default value is [] . :acknowledger - a function that generates ack fields of the sent Broadway.Message.t() . This function receives the acknowledger data and the from field and it must return the acknowledger tuple. The typespec of this function is: data :: term(), from :: {pid(), term()} -&gt; {module(), ack_ref :: term(), ack_data :: term()} :batch_mode - when set to :flush , the batch the message is in is immediately delivered. When set to :bulk , batch is delivered when its size or timeout is reached. The default value is :bulk . Examples For example, in your tests, you may do: ref = Broadway . test_message ( broadway , 1 ) assert_receive { :ack , ^ ref , [ successful ] , [ ] } or if you want to override which acknowledger shall be called, you may do: acknowledger = fn data , ack_ref -&gt; { MyAck , ack_ref , :ok } end Broadway . test_message ( broadway , 1 , acknowledger : acknowledger ) Note that messages sent using this function will ignore demand and :transform option specified in :producer option in Broadway.start_link/2 .","ref":"Broadway.html#test_message/3","title":"Broadway.test_message/3","type":"function"},{"doc":"Returns the topology details for a pipeline. The stages that have the &quot;concurrency&quot; field indicates a list of processes running with that name prefix. Each process has &quot;name&quot; as prefix plus &quot;_&quot; and the index of 0..(concurrency - 1) , as atom. For example, a producer named MyBroadway.Broadway.Producer with concurrency of 1 represents only a process named MyBroadway.Broadway.Producer_0 . Note that Broadway does not accept multiple producers neither multiple processors. But we choose to keep in a list for simplicity and future proof. Examples iex&gt; Broadway . topology ( MyBroadway ) [ producers : [ %{ name : MyBroadway.Broadway.Producer , concurrency : 1 } ] , processors : [ %{ name : MyBroadway.Broadway.Processor_default , concurrency : 10 , processor_key : :default } ] , batchers : [ %{ batcher_name : MyBroadway.Broadway.Batcher_default , name : MyBroadway.Broadway.BatchProcessor_default , batcher_key : :default , concurrency : 5 } , %{ batcher_name : MyBroadway.Broadway.Batcher_s3 , name : MyBroadway.Broadway.BatchProcessor_s3 , batcher_key : :s3 , concurrency : 3 } ] ]","ref":"Broadway.html#topology/1","title":"Broadway.topology/1","type":"function"},{"doc":"Updates the producer rate limiting of the given pipeline at runtime. Supports the following options (see the :rate_limiting options in the module documentation for more information): :allowed_messages :interval Returns an {:error, reason} tuple if the given broadway pipeline doesn't have rate limiting enabled. Examples Broadway . update_rate_limiting ( broadway , allowed_messages : 100 )","ref":"Broadway.html#update_rate_limiting/2","title":"Broadway.update_rate_limiting/2","type":"function"},{"doc":"","ref":"Broadway.html#t:name/0","title":"Broadway.name/0","type":"type"},{"doc":"Returned by start_link/2 .","ref":"Broadway.html#t:on_start/0","title":"Broadway.on_start/0","type":"type"},{"doc":"A struct used to hold information about a generated batch. An instance of this struct containing the related info will be passed to the Broadway.handle_batch/4 callback of the module implementing the Broadway behaviour.","ref":"Broadway.BatchInfo.html","title":"Broadway.BatchInfo","type":"module"},{"doc":"","ref":"Broadway.BatchInfo.html#t:t/0","title":"Broadway.BatchInfo.t/0","type":"type"},{"doc":"This struct holds all information about a message. A message is first created by the producers. It is then sent downstream and gets updated multiple times, either by a module implementing the Broadway behaviour through the Broadway.handle_message/3 callback or internally by one of the built-in stages of Broadway. Instead of modifying the struct directly, you should use the functions provided by this module to manipulate messages.","ref":"Broadway.Message.html","title":"Broadway.Message","type":"module"},{"doc":"Immediately acknowledges the given message or list of messages. This function can be used to acknowledge a message (or list of messages) immediately without waiting for the rest of the pipeline. Acknowledging a message sets that message's acknowledger to a no-op acknowledger so that it's safe to ack at the end of the pipeline. Returns the updated acked message if a message is passed in, or the updated list of acked messages if a list of messages is passed in.","ref":"Broadway.Message.html#ack_immediately/1","title":"Broadway.Message.ack_immediately/1","type":"function"},{"doc":"Configures the acknowledger of this message. This function calls the Broadway.Acknowledger.configure/3 callback to change the configuration of the acknowledger for the given message . This function can only be called if the acknowledger implements the configure/3 callback. If it doesn't, an error is raised.","ref":"Broadway.Message.html#configure_ack/2","title":"Broadway.Message.configure_ack/2","type":"function"},{"doc":"Mark a message as failed. Failed messages are sent directly to the related acknowledger at the end of this step and therefore they're not forwarded to the next step in the pipeline. Failing a message does not emit any log but it does trigger the Broadway.handle_failed/2 callback.","ref":"Broadway.Message.html#failed/2","title":"Broadway.Message.failed/2","type":"function"},{"doc":"Defines the message batch key. The batch key identifies the batch the message belongs to, within a given batcher. Each batcher then groups batches with the same batch_key , with size of at most batch_size within period batch_timeout . Both batch_size and batch_timeout are managed per batch key, so a batcher is capable of grouping multiple batch keys at the same time, regardless of the concurrency level. If a given batcher has multiple batch processors (concurrency &gt; 1), all messages with the same batch key are routed to the same processor. So different batch keys may run concurrently but the same batch key is always run serially and in the same batcher processor.","ref":"Broadway.Message.html#put_batch_key/2","title":"Broadway.Message.put_batch_key/2","type":"function"},{"doc":"Sets the batching mode for the message. When the mode is :bulk , the batch that the message is in is delivered after the batch size or batch timeout is reached. When the mode is :flush , the batch that the message is in is delivered immediately after processing. Note it doesn't mean the batch contains only a single element but rather that all messages received from the processor are delivered without waiting. The default mode for messages is :bulk .","ref":"Broadway.Message.html#put_batch_mode/2","title":"Broadway.Message.put_batch_mode/2","type":"function"},{"doc":"Defines the target batcher which the message should be forwarded to.","ref":"Broadway.Message.html#put_batcher/2","title":"Broadway.Message.put_batcher/2","type":"function"},{"doc":"Stores the given data in the message. This function is usually used inside the Broadway.handle_message/3 implementation to replace data with new processed data.","ref":"Broadway.Message.html#put_data/2","title":"Broadway.Message.put_data/2","type":"function"},{"doc":"Updates the data in the message. This function is usually used inside the Broadway.handle_message/3 implementation to update data with new processed data.","ref":"Broadway.Message.html#update_data/2","title":"Broadway.Message.update_data/2","type":"function"},{"doc":"","ref":"Broadway.Message.html#t:t/0","title":"Broadway.Message.t/0","type":"type"},{"doc":"A behaviour used to acknowledge that the received messages were successfully processed or failed. When implementing a new connector for Broadway, you should implement this behaviour and consider how the technology you're working with handles message acknowledgement. The ack/3 callback must be implemented in order to notify the origin of the data that a message can be safely removed after been successfully processed and published. In case of failed messages or messages without acknowledgement, depending on the technology chosen, the messages can be either moved back in the queue or, alternatively, moved to a dead-letter queue .","ref":"Broadway.Acknowledger.html","title":"Broadway.Acknowledger","type":"behaviour"},{"doc":"Invoked to acknowledge successful and failed messages. ack_ref is a term that uniquely identifies how messages should be grouped and sent for acknowledgement. Imagine you have a scenario where messages are coming from different producers. Broadway will use this information to correctly identify the acknowledger and pass it among with the messages so you can properly communicate with the source of the data for acknowledgement. successful is the list of messages that were successfully processed and published. failed is the list of messages that, for some reason, could not be processed or published.","ref":"Broadway.Acknowledger.html#c:ack/3","title":"Broadway.Acknowledger.ack/3","type":"callback"},{"doc":"Configures the acknowledger with new options . Every acknowledger can decide how to incorporate the given options into its ack_data . The ack_data is the current acknowledger's data. The return value of this function is {:ok, new_ack_data} where new_ack_data is the updated data for the acknowledger. Note that options are different for every acknowledger, as the acknowledger is what specifies what are the supported options. Check the documentation for the acknowledger you're using to see the supported options.","ref":"Broadway.Acknowledger.html#c:configure/3","title":"Broadway.Acknowledger.configure/3","type":"callback"},{"doc":"A simple acknowledger that sends a message back to a caller. It must be stored as: acknowledger : { Broadway.CallerAcknowledger , { pid , ref } , term } The second element is a tuple with the pid to receive the messages and a unique identifier (usually a reference). The third element, which is per message, is ignored. It sends a message in the format: { :ack , ref , successful_messages , failed_messages } If Broadway.Message.configure_ack/2 is called on a message that uses this acknowledger, then the following message is sent: { :configure , ref , options }","ref":"Broadway.CallerAcknowledger.html","title":"Broadway.CallerAcknowledger","type":"module"},{"doc":"An acknowledger that does nothing. Set automatically on messages that have been acked immediately via Broadway.Message.ack_immediately/1 .","ref":"Broadway.NoopAcknowledger.html","title":"Broadway.NoopAcknowledger","type":"module"},{"doc":"A producer that does nothing, used mostly for testing. See &quot;Testing&quot; section in Broadway module documentation for more information.","ref":"Broadway.DummyProducer.html","title":"Broadway.DummyProducer","type":"module"},{"doc":"A Broadway producer is a GenStage producer that emits Broadway.Message structs as events. The Broadway.Producer is declared in a Broadway topology via the :module option: producer : [ module : { MyProducer , options } ] Once declared, MyProducer is expected to implement and behave as a GenStage producer. When Broadway starts, the GenStage.init/1 callback will be invoked with the given options. If options is a keyword list, a :broadway option is injected into such keyword list containing the configuration for the complete Broadway topology. For example, you can use options[:broadway][:name] to uniquely identify the topology, allowing you to write terms to persistent_term or ets . The :broadway configuration also has an :index key which is the index of the producer in its supervision tree (starting from 0). This allows a features such having even producers connect to some server while odd producers connect to another. If options is any other term, it is passed as is to the GenStage.init/1 callback as is. All other functions behave precisely as in GenStage with the requirements that all emitted events must be Broadway.Message structs. Optional callbacks A Broadway.Producer can implement two optional Broadway callbacks: prepare_for_start/2 and prepare_for_draining/1 , which are useful for booting up and shutting down Broadway topologies respectively.","ref":"Broadway.Producer.html","title":"Broadway.Producer","type":"behaviour"},{"doc":"Invoked by the terminator right before Broadway starts draining in-flight messages during shutdown. This callback should be implemented by producers that need to do additional work before shutting down. That includes active producers like RabbitMQ that must ask the data provider to stop sending messages. It will be invoked for each producer stage. state is the current state of the producer.","ref":"Broadway.Producer.html#c:prepare_for_draining/1","title":"Broadway.Producer.prepare_for_draining/1","type":"callback"},{"doc":"Invoked once by Broadway during Broadway.start_link/2 . The goal of this callback is to manipulate the general topology options, if necessary at all, and introduce any new child specs that will be started before the producers supervisor in Broadway's supervision tree. Broadway's supervision tree is a rest_for_one supervisor (see the documentation for Supervisor ), which means that if the children returned from this callback crash they will bring down the rest of the pipeline before being restarted. This callback is guaranteed to be invoked inside the Broadway main process. module is the Broadway module passed as the first argument to Broadway.start_link/2 . options is all of Broadway topology options passed as the second argument to Broadway.start_link/2 . The return value of this callback is a tuple {child_specs, options} . child_specs is the list of child specs to be started under Broadway's supervision tree. updated_options is a potentially-updated list of Broadway options that will be used instead of the ones passed to Broadway.start_link/2 . This can be used to modify the characteristics of the Broadway topology to accommodated for the children started here. Examples defmodule MyProducer do @behaviour Broadway.Producer # other callbacks... @impl true def prepare_for_start ( _module , broadway_options ) do children = [ { DynamicSupervisor , strategy : :one_for_one , name : MyApp.DynamicSupervisor } ] { children , broadway_options } end end","ref":"Broadway.Producer.html#c:prepare_for_start/2","title":"Broadway.Producer.prepare_for_start/2","type":"callback"},{"doc":"Introduction Broadway is a library for building concurrent and multi-stage data ingestion and data processing pipelines with Elixir. Broadway pipelines are concurrent and robust, thanks to the Erlang VM and its actors. It features: Back-pressure Automatic acknowledgements at the end of the pipeline Batching Fault tolerance Graceful shutdown Built-in testing Custom failure handling Ordering and partitioning Rate-limiting Metrics","ref":"introduction.html","title":"Introduction","type":"extras"},{"doc":"Currently we officially support four Broadway producers: Amazon SQS: Source - Guide Apache Kafka: Source - Guide Google Cloud Pub/Sub: Source - Guide RabbitMQ: Source - Guide The guides links above will help you get started with your adapter of choice. For API reference, you can check out the Broadway module.","ref":"introduction.html#official-producers","title":"Introduction - Official Producers","type":"extras"},{"doc":"For those interested in rolling their own Broadway Producers (which we actively encourage!), we recommend using the OffBroadway namespace, mirroring the Off-Broadway theaters . For example, if you want to publish your own integration with Amazon SQS, you can package it as off_broadway_sqs , which uses the OffBroadway.SQS namespace. The following Off-Broadway libraries are available (feel free to send a PR adding your own in alphabetical order): off_broadway_kafka : Guide off_broadway_redis : Guide","ref":"introduction.html#non-official-off-broadway-producers","title":"Introduction - Non-official (Off-Broadway) Producers","type":"extras"},{"doc":"Amazon SQS Amazon Simple Queue Service (SQS) is a highly scalable distributed message queuing service provided by Amazon.com. AWS SQS offers two types of message queues: Standard Nearly unlimited throughput Best-effort ordering At-least-once delivery FIFO Limited number of transactions per second (TPS). See Amazon SQS FIFO developer guide for more information on limits. Order in which messages are sent/received is strictly preserved Exactly-once delivery Broadway can work seamlessly with both, Standard and FIFO queues.","ref":"amazon-sqs.html","title":"Amazon SQS","type":"extras"},{"doc":"In order to use Broadway with SQS, we need to: Create a SQS queue (or use an existing one) Configure our Elixir project to use Broadway Define your pipeline configuration Implement Broadway callbacks Run the Broadway pipeline Tuning the configuration (Optional)","ref":"amazon-sqs.html#getting-started","title":"Amazon SQS - Getting Started","type":"extras"},{"doc":"Amazon provides a comprehensive Step-by-step Guide on creating SQS queues. In case you don't have an AWS account and want to test Broadway locally, use can easily install ElasticMQ , which is a message queue system that offers a SQS-compatible query interface.","ref":"amazon-sqs.html#create-a-sqs-queue","title":"Amazon SQS - Create a SQS queue","type":"extras"},{"doc":"In this guide we're going to use BroadwaySQS , which is a Broadway SQS Connector provided by Dashbit . Starting a new project If you plan to start a new project, just run: $ mix new my_app --sup The --sup flag instructs Elixir to generate an application with a supervision tree. Setting up dependencies Add :broadway_sqs to the list of dependencies in mix.exs along the HTTP client of your choice (defaults to :hackney ): def deps do [ ... { :broadway_sqs , &quot;~&gt; 0.7&quot; } , { :hackney , &quot;~&gt; 1.9&quot; } , ] end Don't forget to check for the latest version of dependencies.","ref":"amazon-sqs.html#configure-the-project","title":"Amazon SQS - Configure the project","type":"extras"},{"doc":"Broadway is a process-based behaviour and to define a Broadway pipeline, we need to define three functions: start_link/1 , handle_message/3 and handle_batch/4 . We will cover start_link/1 in this section and the handle_ callbacks in the next one. Similar to other process-based behaviour, start_link/1 simply delegates to Broadway.start_link/2 , which should define the producers, processors, and batchers in the Broadway pipeline. Assuming we want to consume messages from a queue called my_queue , the minimal configuration would be: defmodule MyBroadway do use Broadway alias Broadway.Message def start_link ( _opts ) do Broadway . start_link ( __MODULE__ , name : __MODULE__ , producer : [ module : { BroadwaySQS.Producer , queue_url : &quot;https://us-east-2.queue.amazonaws.com/100000000001/my_queue&quot; } ] , processors : [ default : [ ] ] , batchers : [ default : [ batch_size : 10 , batch_timeout : 2000 ] ] ) end ... callbacks ... end The above configuration also assumes that you have the AWS credentials set up in your environment, for instance, by having the AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY environment variables set. If that's not the case, you will need to pass that information to the client so it can properly connect to the AWS servers. Here is how you can do it: ... producer : [ module : { BroadwaySQS.Producer , queue_url : &quot;https://us-east-2.queue.amazonaws.com/100000000001/my_queue&quot; , config : [ access_key_id : &quot;YOUR_AWS_ACCESS_KEY_ID&quot; , secret_access_key : &quot;YOUR_AWS_SECRET_ACCESS_KEY&quot; ] } ] ... For a full list of options for BroadwaySQS.Producer , please see BroadwaySQS documentation. For general information about setting up Broadway, see Broadway module docs as well as Broadway.start_link/2 . Note: Even though batching is optional since Broadway v0.2, we recommend that all Amazon SQS pipelines have at least a default batcher. This lets you control the exact batch size and frequency that messages are acknowledged to Amazon SQS, often leading to pipelines that are more cost and time efficient.","ref":"amazon-sqs.html#define-the-pipeline-configuration","title":"Amazon SQS - Define the pipeline configuration","type":"extras"},{"doc":"In order to process incoming messages, we need to implement the required callbacks. For the sake of simplicity, we're considering that all messages received from the queue are just numbers: defmodule MyBroadway do use Broadway alias Broadway.Message ... start_link ... @impl true def handle_message ( _ , % Message { data : data } = message , _ ) do message |&gt; Message . update_data ( fn data -&gt; data * data end ) end @impl true def handle_batch ( _ , messages , _ , _ ) do list = messages |&gt; Enum . map ( fn e -&gt; e . data end ) IO . inspect ( list , label : &quot;Got batch of finished jobs from processors, sending ACKs to SQS as a batch.&quot; ) messages end end We are not doing anything fancy here, but it should be enough for our purpose. First we update the message's data individually inside handle_message/3 and then we print each batch inside handle_batch/4 . For more information, see Broadway.handle_message/3 and Broadway.handle_batch/4 .","ref":"amazon-sqs.html#implement-broadway-callbacks","title":"Amazon SQS - Implement Broadway callbacks","type":"extras"},{"doc":"To run your Broadway pipeline, you just need to add as a child in a supervision tree. Most applications have a supervision tree defined at lib/my_app/application.ex . You can add Broadway as a child to a supervisor as follows: children = [ { MyBroadway , [ ] } ] Supervisor . start_link ( children , strategy : :one_for_one ) Now the Broadway pipeline should be started when your application starts. Also, if your Broadway has any dependency (for example, it needs to talk to the database), make sure that Broadway is listed after its dependencies in the supervision tree.","ref":"amazon-sqs.html#run-the-broadway-pipeline","title":"Amazon SQS - Run the Broadway pipeline","type":"extras"},{"doc":"Some of the configuration options available for Broadway come already with a &quot;reasonable&quot; default value. However those values might not suit your requirements. Depending on the number of messages you get, how much processing they need and how much IO work is going to take place, you might need completely different values to optimize the flow of your pipeline. The concurrency option available for every set of producers, processors and batchers, among with max_demand , batch_size , and batch_timeout can give you a great deal of flexibility. The concurrency option controls the concurrency level in each layer of the pipeline. See the notes on Producer concurrency and Batcher concurrency for details. Here's an example on how you could tune them according to your needs. defmodule MyBroadway do use Broadway def start_link ( _opts ) do Broadway . start_link ( __MODULE__ , name : __MODULE__ , producer : [ ... concurrency : 10 , ] , processors : [ default : [ concurrency : 100 , max_demand : 1 , ] ] , batchers : [ default : [ batch_size : 10 , concurrency : 10 , ] ] ) end ... callbacks ... end In order to get a good set of configurations for your pipeline, it's important to respect the limitations of the servers you're running, as well as the limitations of the services you're providing/consuming data to/from. Broadway comes with telemetry, so you can measure your pipeline and help ensure your changes are effective.","ref":"amazon-sqs.html#tuning-the-configuration","title":"Amazon SQS - Tuning the configuration","type":"extras"},{"doc":"Apache Kafka Kafka is a distributed streaming platform that has three key capabilities: Publish and subscribe to streams of records Store streams of records in a fault-tolerant durable way Process streams of records as they occur","ref":"apache-kafka.html","title":"Apache Kafka","type":"extras"},{"doc":"In order to use Broadway with Kafka, we need to: Create a stream of records (or use an existing one) Configure your Elixir project to use Broadway Define your pipeline configuration Implement Broadway callbacks Run the Broadway pipeline","ref":"apache-kafka.html#getting-started","title":"Apache Kafka - Getting Started","type":"extras"},{"doc":"In case you don't have Kafka installed yet, please follow the instructions on Kafka's Quickstart for a clean installation. After initializing Kafka, you can create a new stream by running: $ kafka-topics --create --zookeeper localhost:2181 --partitions 3 --topic test","ref":"apache-kafka.html#create-a-stream-of-records-or-use-an-existing-one","title":"Apache Kafka - Create a stream of records (or use an existing one)","type":"extras"},{"doc":"This guide describes the steps necessary to integrate Broadway with Kafka using BroadwayKafka , which is a Broadway Kafka Connector provided by Dashbit . BroadwayKafka can subscribe to one or more topics and process streams of records using Kafka's Consumer API . Each GenStage producer initialized by BroadwayKafka will be available as a consumer, all registered using the same self-labeled consumer group . Each record published to a topic/partition will be delivered to one consumer instance within each consumer group. Bear in mind that a topic/partition can be assigned to any consumer instance that has been subscribed using the same consumer group, i.e, any Broadway instance or application running on any machine connected to the Kafka cluster. Starting a new project Create a new project running: $ mix new my_app --sup The --sup flag instructs Elixir to generate an application with a supervision tree. Setting up dependencies Add :broadway_kafka to the list of dependencies in mix.exs : def deps do [ ... { :broadway_kafka , &quot;~&gt; 0.3&quot; } ] end Don't forget to check for the latest version of dependencies.","ref":"apache-kafka.html#configure-your-elixir-project-to-use-broadway","title":"Apache Kafka - Configure your Elixir project to use Broadway","type":"extras"},{"doc":"Broadway is a process-based behaviour and to define a Broadway pipeline, we need to define three functions: start_link/1 , handle_message/3 and optionally handle_batch/4 . We will cover start_link/1 in this section and the handle_ callbacks in the next one. Similar to other process-based behaviours, start_link/1 simply delegates to Broadway.start_link/2 , which should define the producers, processors, and batchers in the Broadway pipeline. Assuming we want to consume messages from a topic called test , one possible configuration would be: defmodule MyBroadway do use Broadway alias Broadway.Message def start_link ( _opts ) do Broadway . start_link ( __MODULE__ , name : __MODULE__ , producer : [ module : { BroadwayKafka.Producer , [ hosts : [ localhost : 9092 ] , group_id : &quot;group_1&quot; , topics : [ &quot;test&quot; ] ] } , concurrency : 1 ] , processors : [ default : [ concurrency : 10 ] ] , batchers : [ default : [ batch_size : 100 , batch_timeout : 200 , concurrency : 10 ] ] ) end ... callbacks ... end Note : Pipelines built on top of BroadwayKafka are automatically partitioned. So even though there are multiple processes (stages), these processes will preserve Kafka's ordering semantics when it comes to topics/partitions. Internally, this is achieved by making sure all messages from the same topic/partition will always be forwarded to the same processor and batch processor. For a full list of options for BroadwayKafka.Producer , refer to the official BroadwayKafka documentation. For general information about setting up Broadway, see Broadway module docs as well as Broadway.start_link/2 .","ref":"apache-kafka.html#define-the-pipeline-configuration","title":"Apache Kafka - Define the pipeline configuration","type":"extras"},{"doc":"In order to process incoming messages, we need to implement the required callbacks. For the sake of simplicity, we're considering that all messages received from the topic are just numbers: defmodule MyBroadway do use Broadway alias Broadway.Message ... start_link ... @impl true def handle_message ( _ , message , _ ) do message |&gt; Message . update_data ( fn data -&gt; { data , String . to_integer ( data ) * 2 } end ) end @impl true def handle_batch ( _ , messages , _ , _ ) do list = messages |&gt; Enum . map ( fn e -&gt; e . data end ) IO . inspect ( list , label : &quot;Got batch&quot; ) messages end end We are not doing anything fancy here, but it should be enough for our purpose. First, we update the message's data individually inside handle_message/3 and then we print each batch inside handle_batch/4 . For more information, see Broadway.handle_message/3 and Broadway.handle_batch/4 . Note: Since Broadway v0.2, batching is optional. In case you don't need to group messages as batches for further processing/publishing, you can remove the :batchers configuration along with the handle_batch/4 callback.","ref":"apache-kafka.html#implement-broadway-callbacks","title":"Apache Kafka - Implement Broadway callbacks","type":"extras"},{"doc":"To run your Broadway pipeline, you just need to add as a child in a supervision tree. Most applications have a supervision tree defined at lib/my_app/application.ex . You can add Broadway as a child to a supervisor as follows: children = [ { MyBroadway , [ ] } ] Supervisor . start_link ( children , strategy : :one_for_one ) Now the Broadway pipeline should be started when your application starts. Also, if your Broadway has any dependency (for example, it needs to talk to the database), make sure that Broadway is listed after its dependencies in the supervision tree. You can now test your pipeline by entering an iex session: $ iex -S mix If everything went fine, you should see lots of info log messages like this one coming from the :brod supervisors: 15 : 14 : 04.356 [ info ] [ supervisor : { :local , :brod_sup } , started : [ pid : # PID &lt; 0 . 251 . 0 &gt; , id : :test_client , mfargs : { :brod_client , :start_link , [ [ localhost : 9092 ] , :test_client , [ ] ] } , restart_type : { :permanent , 10 } , shutdown : 5000 , child_type : :worker ] ] Brod is the client that BroadwayKafka uses under the hood to communicate with Kafka. Sending messages to Kafka Finally, we can send some sample messages to Kafka using using :brod with the following snippet: topic = &quot;test&quot; client_id = :my_client hosts = [ localhost : 9092 ] :ok = :brod . start_client ( hosts , client_id , _client_config = [ ] ) :ok = :brod . start_producer ( client_id , topic , _producer_config = [ ] ) Enum . each ( 1 .. 1000 , fn i -&gt; partition = rem ( i , 3 ) :ok = :brod . produce_sync ( client_id , topic , partition , _key = &quot;&quot; , &quot; \#{ i } &quot; ) end ) You should see the output showing the generated batches: Got batch : [ { &quot;2&quot; , 4 } , { &quot;5&quot; , 10 } , { &quot;8&quot; , 16 } , { &quot;11&quot; , 22 } , { &quot;14&quot; , 28 } , ... ] Got batch : [ { &quot;3&quot; , 6 } , { &quot;6&quot; , 12 } , { &quot;9&quot; , 18 } , { &quot;12&quot; , 24 } , { &quot;15&quot; , 30 } , ... ]","ref":"apache-kafka.html#run-the-broadway-pipeline","title":"Apache Kafka - Run the Broadway pipeline","type":"extras"},{"doc":"Some of the configuration options available for Broadway come already with a &quot;reasonable&quot; default value. However, those values might not suit your requirements. Depending on the number of records you get, how much processing they need and how much IO work is going to take place, you might need completely different values to optimize the flow of your pipeline. The concurrency option available for every set of producers, processors and batchers, along with batch_size and batch_timeout can give you a great deal of flexibility. See the notes on Producer concurrency and Batcher concurrency for details. By setting the concurrency option, you define the number of concurrent process that will be started by Broadway, allowing you to have full control over the concurrency level in each layer of the pipeline. Keep in mind that since the concurrency model provided by Kafka is based on partitioning , in order to take full advantage of this model, you need to set the concurrency option for your processors and batchers accordingly. Having less concurrency than topic/partitions assigned will result in individual processors handling more than one partition, decreasing the overall level of concurrency. Therefore, if you want to always be able to process messages at maximum concurrency (assuming you have enough resources to do it), you should increase the concurrency up front to make sure you have enough processors to handle the extra records received from new partitions assigned. Note : Even if you don't plan to add more partitions to a Kafka topic, your pipeline can still receive more assignments than planned. For instance, if another consumer crashes, the server will reassign all its topic/partition to other available consumers, including any Broadway producer subscribed to the same topic. There are other options that you may want to take a closer look when tuning your configuration. The :max_bytes option, for instance, belongs to the :fetch_config group and defines the maximum amount of data to be fetched at a time from a single partition. The default is 1048576 (1 MiB). Setting greater values can improve throughput at the cost of more memory consumption. For more information and other fetch options, please refer to the &quot;Fetch config options&quot; in the official BroadwayKafka documentation. Other two important options are :offset_commit_interval_seconds and :offset_commit_on_ack . Both belong to the main configuration and they can make a huge impact on performance. The :offset_commit_interval_seconds defines the time interval between two OffsetCommitRequest messages. The default is 5s. The :offset_commit_on_ack , when set to true , tells Broadway to send an OffsetCommitRequest immediately after each acknowledgemnt, bypassing any interval defined in :offset_commit_interval_seconds . Setting this option to false can increase performance since any commit requests will start respecting the :offset_commit_interval_seconds option. This will usually result in fewer requests to be sent to the server. However, setting long commit intervals might lead to a large number of duplicated records to be processed after a server restart or connection loss. If that's the case, make sure your logic is idempotent when consuming records to avoid inconsistencies. Also, bear in mind that the negative performance impact might be insignificant if you're using batchers since only one commit request will be performed per batch. As a basic rule, always take into account the values of batch_size and batch_timeout whenever you're tuning :offset_commit_interval_seconds and :offset_commit_on_ack","ref":"apache-kafka.html#tuning-the-configuration","title":"Apache Kafka - Tuning the configuration","type":"extras"},{"doc":"broadway_kafka never stops the flow of the stream, i.e. it will always ack the messages even when they fail. Unlike queue-based connectors, where you can mark a single message as failed. In Kafka that's not possible due to its single offset per topic/partition ack strategy. If you want to reprocess failed messages, you need to roll your own strategy. A possible way to do that is to implement handle_failed/2 and send failed messages to a separated stream or queue for later processing.","ref":"apache-kafka.html#handling-failed-messages","title":"Apache Kafka - Handling failed messages","type":"extras"},{"doc":"Google Cloud Pub/Sub Cloud Pub/Sub is a fully-managed real-time messaging service provided by Google.","ref":"google-cloud-pubsub.html","title":"Google Cloud Pub/Sub","type":"extras"},{"doc":"In order to use Broadway with Cloud Pub/Sub you need to: Setup a Cloud Pub/Sub project Configure your Elixir project to use Broadway Define your pipeline configuration Implement Broadway callbacks Run the Broadway pipeline Tune the configuration (Optional) If you are just getting familiar with Google Pub/Sub, refer to the documentation to get started. Instead of testing against a live environment, you may also consider using the emulator to simulate integrating with Cloud Pub/Sub. If you have an existing project, topic, subscription, and credentials, you can skip step 1 and jump to Configure the project section.","ref":"google-cloud-pubsub.html#getting-started","title":"Google Cloud Pub/Sub - Getting Started","type":"extras"},{"doc":"In this tutorial we'll use the gcloud command-line tool to set everything up in Google Cloud. Alternatively, you can roughly follow this guide by using Cloud Console . To install gcloud follow the documentation . If you are on macOS you may consider installing it with Homebrew: $ brew install --cask google-cloud-sdk Now, authenticate the CLI: $ gcloud auth login Then, create a new project: $ gcloud projects create test-pubsub A new topic: $ gcloud pubsub topics create test-topic --project test-pubsub Created topic [projects/test-pubsub/topics/test-topic]. Note: If you run this command immediately after creating a new Google Cloud project, you may receive an error indicating that your project's organization policy is still being provisioned. Just wait a couple minutes and try again. And a new subscription: $ gcloud pubsub subscriptions create test-subscription --project test-pubsub --topic test-topic Created subscription [projects/test-pubsub/subscriptions/test-subscription]. We also need a service account , an IAM policy, as well as API credentials in order to programmatically work with the service. First, let's create the service account: $ gcloud iam service-accounts create test-account --project test-pubsub Created service account [test-account]. Then the policy. For simplicity we add the general role roles/editor , but make sure to examine the available roles and choose the one that best suits your use case: $ gcloud projects add-iam-policy-binding test-pubsub \\ --member serviceAccount:test-account@test-pubsub.iam.gserviceaccount.com \\ --role roles/editor Updated IAM policy for project [test-pubsub]. (...) And now the credentials: $ gcloud iam service-accounts keys create credentials.json --iam-account=test-account@test-pubsub.iam.gserviceaccount.com created key [xxx] of type [json] as [key] for [test-account@test-pubsub.iam.gserviceaccount.com] This command generated a credentials.json file which will be useful later. Note, the IAM account pattern is &lt;account&gt;@&lt;project&gt;.iam.gserviceaccount.com . Run gcloud iam service-accounts list --project test-pubsub to see all service accounts associated with the given project. Finally, we need to enable Pub/Sub for our project: $ gcloud services enable pubsub --project test-pubsub Operation &quot;operations/xxx&quot; finished successfully.","ref":"google-cloud-pubsub.html#setup-cloud-pub-sub-project","title":"Google Cloud Pub/Sub - Setup Cloud Pub/Sub project","type":"extras"},{"doc":"In this guide we're going to use BroadwayCloudPubSub , which is a Broadway Cloud Pub/Sub Connector provided by Dashbit . Starting a new project If you plan to start a new project, just run: $ mix new my_app --sup The --sup flag instructs Elixir to generate an application with a supervision tree. Setting up dependencies Add :broadway_cloud_pub_sub to the list of dependencies in mix.exs , along with the Google Cloud authentication library of your choice (defaults to :goth ): defp deps ( ) do [ ... { :broadway_cloud_pub_sub , &quot;~&gt; 0.7&quot; } , { :goth , &quot;~&gt; 1.0&quot; } ] end Don't forget to check for the latest version of dependencies.","ref":"google-cloud-pubsub.html#configure-the-project","title":"Google Cloud Pub/Sub - Configure the project","type":"extras"},{"doc":"Broadway is a process-based behaviour and to define a Broadway pipeline, we need to define three functions: start_link/1 , handle_message/3 and handle_batch/4 . We will cover start_link/1 in this section and the handle_ callbacks in the next one. Similar to other process-based behaviour, start_link/1 simply delegates to Broadway.start_link/2 , which should define the producers, processors, and batchers in the Broadway pipeline. Assuming we want to consume messages from the test-subscription , the minimal configuration would be: defmodule MyBroadway do use Broadway alias Broadway.Message def start_link ( _opts ) do Broadway . start_link ( __MODULE__ , name : __MODULE__ , producer : [ module : { BroadwayCloudPubSub.Producer , subscription : &quot;projects/test-pubsub/subscriptions/test-subscription&quot; } ] , processors : [ default : [ ] ] , batchers : [ default : [ batch_size : 10 , batch_timeout : 2_000 ] ] ) end ... callbacks ... end For a full list of options for BroadwayCloudPubSub.Producer , please see the documentation . For general information about setting up Broadway, see Broadway module docs as well as Broadway.start_link/2 . Note: Even though batching is optional since Broadway v0.2, we recommend all Cloud Pub/Sub pipelines to have at least a default batcher, as that allows you to control the exact batch size and frequency that messages are acknowledged to Cloud Pub/Sub, which often leads to pipelines that are more cost and time efficient.","ref":"google-cloud-pubsub.html#define-the-pipeline-configuration","title":"Google Cloud Pub/Sub - Define the pipeline configuration","type":"extras"},{"doc":"In order to process incoming messages, we need to implement the required callbacks. For the sake of simplicity, we're considering that all messages received from the queue are strings and our processor calls String.upcase/1 on them: defmodule MyBroadway do use Broadway alias Broadway.Message ... start_link ... def handle_message ( _ , % Message { data : data } = message , _ ) do message |&gt; Message . update_data ( fn data -&gt; String . upcase ( data ) end ) end def handle_batch ( _ , messages , _ , _ ) do list = messages |&gt; Enum . map ( fn e -&gt; e . data end ) IO . inspect ( list , label : &quot;Got batch of finished jobs from processors, sending ACKs to Pub/Sub as a batch.&quot; ) messages end end We are not doing anything fancy here, but it should be enough for our purpose. First we update the message's data individually inside handle_message/3 and then we print each batch inside handle_batch/4 . For more information, see Broadway.handle_message/3 and Broadway.handle_batch/4 .","ref":"google-cloud-pubsub.html#implement-broadway-callbacks","title":"Google Cloud Pub/Sub - Implement Broadway callbacks","type":"extras"},{"doc":"To run your Broadway pipeline, you need to add it as a child in a supervision tree. Most applications have a supervision tree defined at lib/my_app/application.ex . You can add Broadway as a child to a supervisor as follows: children = [ { MyBroadway , [ ] } ] Supervisor . start_link ( children , strategy : :one_for_one ) The final step is to configure credentials. You can set the following environment variable: export GOOGLE_APPLICATION_CREDENTIALS = / path / to / credentials . json See Goth documentation for alternative ways of authenticating with the API. Now the Broadway pipeline should be started when your application starts. Also, if your Broadway pipeline has any dependency (for example, it needs to talk to the database), make sure that it is listed after its dependencies in the supervision tree. If you followed the previous section about setting the project with gcloud , you can now test the the pipeline. In one terminal tab start the application: $ iex -S mix And in another tab, send a couple of test messages to Pub/Sub: $ gcloud pubsub topics publish projects/test-pubsub/topics/test-topic --message &quot;test 1&quot; messageIds: - &#39;651428033718119&#39; gcloud pubsub topics publish projects/test-pubsub/topics/test-topic --message &quot;test 2&quot; messageIds: - &#39;651427034966696&#39; Now, In the first tab, you should see output similar to: Got batch of finished jobs from processors , sending ACKs to Pub / Sub as a batch . : [ &quot;TEST 1&quot; , &quot;TEST 2&quot; ]","ref":"google-cloud-pubsub.html#run-the-broadway-pipeline","title":"Google Cloud Pub/Sub - Run the Broadway pipeline","type":"extras"},{"doc":"Some of the configuration options available for Broadway come already with a &quot;reasonable&quot; default value. However those values might not suit your requirements. Depending on the number of messages you get, how much processing they need and how much IO work is going to take place, you might need completely different values to optimize the flow of your pipeline. The concurrency option available for every set of producers, processors and batchers, among with max_demand , batch_size , and batch_timeout can give you a great deal of flexibility. The concurrency option controls the concurrency level in each layer of the pipeline. See the notes on Producer concurrency and Batcher concurrency for details. Here's an example on how you could tune them according to your needs. defmodule MyBroadway do use Broadway def start_link ( _opts ) do Broadway . start_link ( __MODULE__ , name : __MODULE__ , producer : [ ... concurrency : 10 , ] , processors : [ default : [ concurrency : 100 , max_demand : 1 , ] ] , batchers : [ default : [ batch_size : 10 , concurrency : 10 , ] ] ) end ... callbacks ... end In order to get a good set of configurations for your pipeline, it's important to respect the limitations of the servers you're running, as well as the limitations of the services you're providing/consuming data to/from. Broadway comes with telemetry, so you can measure your pipeline and help ensure your changes are effective.","ref":"google-cloud-pubsub.html#tuning-the-configuration","title":"Google Cloud Pub/Sub - Tuning the configuration","type":"extras"},{"doc":"RabbitMQ RabbitMQ is an open source message broker designed to be highly scalable and distributed. It supports multiple protocols including the Advanced Message Queuing Protocol (AMQP).","ref":"rabbitmq.html","title":"RabbitMQ","type":"extras"},{"doc":"In order to use Broadway with RabbitMQ, we need to: Create a queue (or use an existing one) Configure our Elixir project to use Broadway Define your pipeline configuration Implement Broadway callbacks Run the Broadway pipeline Tuning the configuration (Optional) In case you want to work with an existing queue, you can skip step 1 and jump to Configure the project . Note: BroadwayRabbitMQ does not automatically create any queue. If you configure a pipeline with a non-existent queue, the producers will crash, bringing down the pipeline.","ref":"rabbitmq.html#getting-started","title":"RabbitMQ - Getting Started","type":"extras"},{"doc":"RabbitMQ runs on many operating systems. Please see Downloading and Installing RabbitMQ for further information. Also, make sure you have the Management plugin enabled, which ships with the command line tool, rabbitmqadmin . After successfully installing RabbitMQ, you can declare a new queue with the following command: $ rabbitmqadmin declare queue name=my_queue durable=true You can list all declared queues to see our the one we've just created: $ rabbitmqctl list_queues Timeout: 60.0 seconds ... Listing queues for vhost / ... name messages my_queue 0","ref":"rabbitmq.html#create-a-queue","title":"RabbitMQ - Create a queue","type":"extras"},{"doc":"In this guide, we're going to use BroadwayRabbitMQ , which is a Broadway RabbitMQ Connector provided by Dashbit . Starting a new project If you're creating a new project, run: $ mix new my_app --sup The --sup flag instructs Elixir to generate an application with a supervision tree. Setting up dependencies Add :broadway_rabbitmq to the list of dependencies in mix.exs : def deps do [ ... { :broadway_rabbitmq , &quot;~&gt; 0.7&quot; } , ] end Don't forget to check for the latest version of dependencies.","ref":"rabbitmq.html#configure-the-project","title":"RabbitMQ - Configure the project","type":"extras"},{"doc":"Broadway is a process-based behaviour and to define a Broadway pipeline, we need to define three functions: start_link/1 , handle_message/3 and optionally handle_batch/4 . We will cover start_link/1 in this section and the handle_ callbacks in the next one. Similar to other process-based behaviours, start_link/1 simply delegates to Broadway.start_link/2 , which should define the producers, processors, and batchers in the Broadway pipeline. Assuming we want to consume messages from a queue called my_queue , one possible configuration would be: defmodule MyBroadway do use Broadway alias Broadway.Message def start_link ( _opts ) do Broadway . start_link ( __MODULE__ , name : MyBroadway , producer : [ module : { BroadwayRabbitMQ.Producer , queue : &quot;my_queue&quot; , qos : [ prefetch_count : 50 , ] } , concurrency : 1 ] , processors : [ default : [ concurrency : 50 ] ] , batchers : [ default : [ batch_size : 10 , batch_timeout : 1500 , concurrency : 5 ] ] ) end ... callbacks ... end If you're consuming data from an existing broker that requires authorization, you'll need to provide your credentials using the connection option: ... producer : [ module : { BroadwayRabbitMQ.Producer , queue : &quot;my_queue&quot; , connection : [ username : &quot;user&quot; , password : &quot;password&quot; , ] ... } ] ... For the full list of connection options, please see AMQP.Connection.open/1 For general information about setting up Broadway, see Broadway module docs as well as Broadway.start_link/2 .","ref":"rabbitmq.html#define-the-pipeline-configuration","title":"RabbitMQ - Define the pipeline configuration","type":"extras"},{"doc":"In order to process incoming messages, we need to implement the required callbacks. For the sake of simplicity, we're considering that all messages received from the queue are just numbers: defmodule MyBroadway do use Broadway alias Broadway.Message ... start_link ... @impl true def handle_message ( _ , message , _ ) do message |&gt; Message . update_data ( fn data -&gt; { data , String . to_integer ( data ) * 2 } end ) end @impl true def handle_batch ( _ , messages , _ , _ ) do list = messages |&gt; Enum . map ( fn e -&gt; e . data end ) IO . inspect ( list , label : &quot;Got batch&quot; ) messages end end We are not doing anything fancy here, but it should be enough for our purpose. First, we update the message's data individually inside handle_message/3 and then we print each batch inside handle_batch/4 . For more information, see Broadway.handle_message/3 and Broadway.handle_batch/4 . Note: Since Broadway v0.2, batching is optional. In case you don't need to group messages as batches for further processing/publishing, you can remove the :batchers configuration along with the handle_batch/4 callback. This is perfectly fine for RabbitMQ, where messages are acknowledged individually and never as a batch.","ref":"rabbitmq.html#implement-broadway-callbacks","title":"RabbitMQ - Implement Broadway callbacks","type":"extras"},{"doc":"To run your Broadway pipeline, you just need to add as a child in a supervision tree. Most applications have a supervision tree defined at lib/my_app/application.ex . You can add Broadway as a child to a supervisor as follows: children = [ { MyBroadway , [ ] } ] Supervisor . start_link ( children , strategy : :one_for_one ) Now the Broadway pipeline should be started when your application starts. Also, if your Broadway has any dependency (for example, it needs to talk to the database), make sure that Broadway is listed after its dependencies in the supervision tree. You can now test your pipeline by entering an iex session: $ iex -S mix If everything went fine, you should see lots of info log messages from the amqp supervisors. If you think that's too verbose and want to do something about it, please take a look at the &quot;Log related to amqp supervisors are too verbose&quot; subsection in the amqp 's Troubleshooting documentation. Finally, let's generate some sample messages to be consumed by Broadway with the following code: { :ok , connection } = AMQP.Connection . open { :ok , channel } = AMQP.Channel . open ( connection ) AMQP.Queue . declare ( channel , &quot;my_queue&quot; , durable : true ) Enum . each ( 1 .. 5000 , fn i -&gt; AMQP.Basic . publish ( channel , &quot;&quot; , &quot;my_queue&quot; , &quot; \#{ i } &quot; ) end ) AMQP.Connection . close ( connection ) You should see the output showing the generated batches: Got batch : [ { &quot;7&quot; , 14 } , { &quot;5&quot; , 10 } , { &quot;8&quot; , 16 } , { &quot;98&quot; , 196 } , { &quot;6&quot; , 12 } , { &quot;97&quot; , 194 } , { &quot;9&quot; , 18 } , { &quot;99&quot; , 198 } , { &quot;10&quot; , 20 } , { &quot;100&quot; , 200 } ] Got batch : [ { &quot;29&quot; , 58 } , { &quot;32&quot; , 64 } , ... ]","ref":"rabbitmq.html#run-the-broadway-pipeline","title":"RabbitMQ - Run the Broadway pipeline","type":"extras"},{"doc":"Some of the configuration options available for Broadway come already with a &quot;reasonable&quot; default value. However, those values might not suit your requirements. Depending on the number of messages you get, how much processing they need and how much IO work is going to take place, you might need completely different values to optimize the flow of your pipeline. The concurrency option available for every set of producers, processors and batchers, among with max_demand , batch_size , and batch_timeout can give you a great deal of flexibility. The concurrency option controls the concurrency level in each layer of the pipeline. See the notes on Producer concurrency and Batcher concurrency for details. Another important option to take into account is the :prefetch_count . RabbitMQ will continually push new messages to Broadway as it receives them. The :prefetch_count setting provides back-pressure by instructing RabbitMQ to limit the number of unacknowledged messages a consumer will have at a given moment . See the &quot;Back-pressure and :prefetch_count&quot; section of the BroadwayRabbitMQ documentation for details. In order to get a good set of configurations for your pipeline, it's important to respect the limitations of the servers you're running, as well as the limitations of the services you're providing/consuming data to/from. Broadway comes with telemetry, so you can measure your pipeline and help ensure your changes are effective.","ref":"rabbitmq.html#tuning-the-configuration","title":"RabbitMQ - Tuning the configuration","type":"extras"},{"doc":"Custom Producers If you want to use Broadway but there is no existing Broadway producer for the technology of your choice, you can integrate any existing GenStage producer into the pipeline with relative ease.","ref":"custom-producers.html","title":"Custom Producers","type":"extras"},{"doc":"In general, producers must generate %Broadway.Message{} structs in order to be processed by Broadway. In case you need to use an existing GenStage producer and you don't want to change its original implementation, you'll have to set the producer's :transformer option to translate the generated events into Broadway messages. In the following example the producer is a regular GenStage, i.e., it produces plain events that cannot be processed by Broadway directly: defmodule Counter do use GenStage def start_link ( number ) do GenStage . start_link ( Counter , number ) end def init ( counter ) do { :producer , counter } end def handle_demand ( demand , counter ) when demand &gt; 0 do events = Enum . to_list ( counter .. counter + demand - 1 ) { :noreply , events , counter + demand } end end By using a transformer, you can tell Broadway to transform all events generated by the producer into proper Broadway messages: defmodule MyBroadway do use Broadway alias Broadway.Message def start_link ( _opts ) do Broadway . start_link ( __MODULE__ , name : __MODULE__ , producer : [ module : { Counter , 1 } , transformer : { __MODULE__ , :transform , [ ] } ] , processors : [ default : [ concurrency : 10 ] ] , batchers : [ default : [ concurrency : 2 , batch_size : 5 ] , ] ) end ... callbacks ... def transform ( event , _opts ) do % Message { data : event , acknowledger : { __MODULE__ , :ack_id , :ack_data } } end def ack ( :ack_id , successful , failed ) do # Write ack code here end end Notice that you need to pass two options to the producer: :module - a tuple representing the GenStage producer as {mod, arg} . Where mod is module that implements the GenStage behaviour and arg the argument that will be given to the init callback of the GenStage. It is very important to note that Broadway will not call the child_spec/1 or start_link/1 function of the producer. That's because Broadway wraps the producer to augment it with extra features. :transformer - a module-function-args tuple that will be invoked inside the producer, for every producer message, that should create a Broadway.Message struct with the data and acknowledger fields. See the Broadway.Acknowledger module for more information on defining and setting up acknowledgements.","ref":"custom-producers.html#example","title":"Custom Producers - Example","type":"extras"},{"doc":"Architecture Broadway's architecture is built on top of GenStage. That means we structure our processing units as independent stages that are responsible for one individual task in the pipeline. By implementing the Broadway behaviour, we define a GenServer process that wraps a Supervisor to manage and own our pipeline.","ref":"architecture.html","title":"Architecture","type":"extras"},{"doc":"[producers] &lt;- pulls data from SQS, RabbitMQ, etc. | | (demand dispatcher) | handle_message/3 and ----------&gt; [processors] prepare_messages/2 run here / \\ / \\ (partition dispatcher) / \\ [batcher] [batcher] &lt;- one for each batcher key | | | | (demand dispatcher) | | handle_batch/4 runs here -&gt; [batch processor][batch processor]","ref":"architecture.html#the-pipeline-model","title":"Architecture - The pipeline model","type":"extras"},{"doc":"Broadway.Producer - A wrapper around the actual producer defined by the user. It serves as the source of the pipeline. Broadway.Processor - This is where messages are processed, e.g. do calculations, convert data into a custom json format etc. Here is where the code from handle_message/3 runs. Broadway.Batcher - Creates batches of messages based on the batcher's key. One batcher for each key will be created. Broadway.BatchProcessor - This is where the code from handle_batch/4 runs.","ref":"architecture.html#internal-stages","title":"Architecture - Internal stages","type":"extras"},{"doc":"Broadway was designed to always go back to a working state in case of failures thanks to the use of supervisors. Our supervision tree is designed as follows: [Broadway GenServer] | | | [Broadway Pipeline Supervisor] / / (:rest_for_one) \\ \\ / | | \\ / | | \\ / | | \\ / | | \\ / | | \\ [ProducerSupervisor] [ProcessorSupervisor] [BatchersSupervisor] [Terminator] (:one_for_one) (:one_for_all) (:one_for_one) / \\ / \\ / \\ / \\ / \\ / \\ / \\ / \\ / \\ / \\ / \\ / \\ [Producer_1] ... [Processor_1] ... [BatcherSupervisor_1] ... (:rest_for_one) / \\ / \\ / \\ [Batcher] [BatchProcessorSupervisor] (:one_for_all) / \\ / \\ / \\ [BatchProcessor_1] ... Both ProcessorSupervisor and ConsumerSupervisor are set with max_restarts to 0. The idea is that if any process fails, we want to restart the rest of the tree. Since Broadway callbacks are stateless, we can handle errors and provide reports without crashing processes. This means that the supervision tree will only shutdown in case of unforeseen errors in Broadway's implementation. The only exception are the producers, which contain external code and are expected to fail. If a producer crashes, it will be restarted by its supervisor without cascading failures until its max restarts is reached. Broadway automatically handles those failures by making processors automatically resubscribe to producers in case of crashes.","ref":"architecture.html#the-supervision-tree","title":"Architecture - The supervision tree","type":"extras"},{"doc":"The cascading failures aspect also provides safe semantics for graceful shutdown. We know that either all processes are running OR they are all being shutdown. Therefore, to gracefully shutdown the supervision tree, a terminator process is activated, which starts the following steps: It notifies the first layer of processors that they should not resubscribe to producers once they exit It tells all producers to no longer accept demand, flush all current events, and then shutdown It then monitors and waits for a confirmation message from batch processors. At this point, the terminator is effectively blocking the supervisor until all events have been processed This triggers a cascade effect where processors notice all of its producers have been cancelled, causing them to flush their own events and cancels the stages downstream, and so on and so on. This happens until batch processors notice all of their producers have been cancelled, effectively notifying the terminator to shutdown, allowing the outer most supervisor to go on and fully terminate all stages, which at this point have flushed all events.","ref":"architecture.html#graceful-shutdowns","title":"Architecture - Graceful shutdowns","type":"extras"}]